bl_info = {"name": "Export 10tons Animation", 
           "author": "10tons Ltd.",
           "version": (0, 7, 6),
           "blender": (2, 73, 0),
           "location": "File > Import-Export",
           "description": "Exports animations to .xml or .anim format.",
           "category": "Import-Export"}

import bpy
import struct

from bpy_extras.io_utils import ImportHelper, ExportHelper
from bpy.props import EnumProperty, BoolProperty

import time, os, re
import mathutils, math, struct

exporter_name = "10tons Animation"
encoding = "iso-8859-1"

class ExitAbort(Exception):
    pass

class Export10tonsAnimation(bpy.types.Operator, ExportHelper):
    bl_idname = "export_scene.the10tons_animation"
    bl_label = "Export " + exporter_name

    filename_ext = ""

    export_format_items = [("ANIM", ".anim", "Fast loading proprietary binary format.", 0), ("XML", ".xml", "Human readable XML based format.", 1)]
    export_format_enum = EnumProperty(name="Format", items=export_format_items)

    export_last_frame_in_range = BoolProperty(name="Export last frame in range", description="", default = True)
    only_export_selected_action_animation_per_object = BoolProperty(name="Only export selected animation per object", description="", default = False)

    def invoke(self, context, event):
        
        objects = list(ob for ob in context.scene.objects if ob.select)
        
        for ob in objects:
            ob = get_armature_to_export_for_object(ob)
            if (ob != None):
                self.filepath = ob.name+"-"+ob.animation_data.action.name
            
        if (len(objects) > 1):
            self.filepath = ""

        context.window_manager.fileselect_add(self)
        return {'RUNNING_MODAL'}

    def execute(self, context):
        if (self.properties.export_format_enum == "XML"):
            self.filename_ext = ".xml"
        elif (self.properties.export_format_enum == "ANIM"):
            self.filename_ext = ".anim"
        else:
            self.filename_ext = ".unknown"
            return
        try:
            path = self.properties.filepath + self.filename_ext
            result = save(path, context, self.properties.export_last_frame_in_range, self.properties.export_format_enum, self.properties.only_export_selected_action_animation_per_object)
            return result
        except ExitAbort as e:
            print("Aborting export. Fix the errors shown above.")
            print("Failed:", e)
            return {'FINISHED'}
			

def save(filepath, context, export_last_frame_in_range, export_format, only_export_selected_action_animation_per_object):
    print("\n\n10tons Animation Exporter.")
   
    if (export_format == "XML"):
        print("Exporting .xml format.")
    elif (export_format == "ANIM"):
        print("Exporting .anim format.")
    else:
        print("Unknown export format. Not doing anything.")
        return {'CANCELLED'}

    previous_phase_start_time = time.clock()
		
    path = os.path.split(filepath)[0]

    scene = context.scene    
    
    objects = [ob for ob in scene.objects if ob.select]
    if not objects:
        objects = scene.objects

    old_frame = scene.frame_current
    
    list_of_exported_objects = []

    for ob in objects:
        ob = get_armature_to_export_for_object(ob)
        if ob == None:
            continue
            
        list_of_exported_objects.append(ob)
   
        
    for ob in list_of_exported_objects:

        old_action = ob.animation_data.action

        list_of_exported_actions = []
        
        if (only_export_selected_action_animation_per_object):
            if (should_export_action(ob, ob.animation_data.action)):
                list_of_exported_actions.append(ob.animation_data.action)
        else:
            for action in bpy.data.actions:
                if (should_export_action(ob, action)):
                    list_of_exported_actions.append(action)
            
        
        print("Exporting rig '"+ob.name+"'")
                
        for action in list_of_exported_actions:
            if (not should_export_action(ob, action)):
                continue
				
            bones = []
            for pb in ob.pose.bones:
                if (should_export_bone(pb)):
                    bones.append(pb)
			
            ob.data.pose_position = 'POSE'                    
            ob.animation_data.action = action
			
            # Set the frame to the last frame to get around a Blender bug with constraints
            scene.frame_set(int(action.frame_range[1]))
            start_frame_number = int(action.frame_range[0])
            end_frame_number = int(action.frame_range[1])
            
            if (export_last_frame_in_range):
                end_frame_number = end_frame_number + 1

            print("Exporting animation '"+action.name+'"')
			
            if (export_format == "XML"):
                if (only_export_selected_action_animation_per_object):
                    filename = filepath
                else:
                    filename = path+"\\"+ob.name+"-"+action.name+".xml"
                save_animation_to_xml(filename, scene, ob.name, action.name, bones, start_frame_number, end_frame_number)
			   
            elif (export_format == "ANIM"):
                if (only_export_selected_action_animation_per_object):
                    filename = filepath
                else:
                    filename = path+"\\"+ob.name+"-"+action.name+".anim"
                save_animation_to_anim(filename, scene, ob.name, action.name, bones, start_frame_number, end_frame_number)
            
        ob.animation_data.action = old_action
        
    scene.frame_set(old_frame)

    print("Exporting took %.2f seconds" % (time.clock() - previous_phase_start_time))
    return {'FINISHED'}

    
def get_armature_to_export_for_object(ob):
    if ob.type == 'MESH':
        arma = ob.find_armature()
        if arma != None:
            ob = arma
            
    if ob.type != 'ARMATURE':
        ob = None
    return ob
	
def should_export_action(ob, action):
    print("Action: " + action.name)

    if action.frame_range[1] <= 1:
        print("Skipping action because frame_range <= 1.")
        return False
    
    # Check that every action group has a corresponding bone in the armature
    action_affects_rig = True
    bone_found = True
    for grp in action.groups:
        bone_found = False
        for b in ob.data.bones:
            if b.name == grp.name:
                bone_found = True
                break
        if not bone_found:        	
            action_affects_rig = False
            print("Bone '" + b.name + "' not found in action '" + action.name + "' channel '" + grp.name)
            break

    if not action_affects_rig:
        print("Action '"+action.name+"' does not have matching bones '"+ob.name+"'. Skipping action.")
        return False
        
    return True
    
        
def save_animation_to_xml(filename, scene, rig_name, action_name, bones, start_frame_number, end_frame_number):
    f = open(filename, "w")
    
    f.write('<?xml version="1.0" encoding="iso-8859-1"?>\n')
    f.write('<root>\n\n')
    
    f.write(' <array id="INFO">\n')
    f.write('  <node id="format" version="0.1.0" />\n')
    f.write('  <node id="animation_name" value="'+action_name+'" />\n')
    f.write('  <node id="frames_per_second" value="'+str(int(scene.render.fps / scene.render.fps_base))+'" />\n')
    f.write('  <node id="skeleton" value="'+rig_name+'" />\n')
    f.write(' </array>\n\n')
    
    f.write(' <array id="BONES">\n')
    for pb in bones:
        f.write('  <node id="'+pb.name+'" />\n')
    f.write(' </array>\n\n')

    
    for frame in range(start_frame_number, end_frame_number):
        scene.frame_set(frame)
        i = 0
        
        f.write(' <array id="FRAME_'+str(frame)+'" >\n')
        
        for pb in bones:
            frame_info = FrameInfo(pb)
            
            f.write('  <node bone_index="'+str(i)+'" ')
            f.write('px="%.5f" ' % frame_info.position.x)
            f.write('py="%.5f" ' % frame_info.position.y)
            f.write('pz="%.5f" ' % frame_info.position.z)
            f.write('qw="%.5f" ' % frame_info.quaternion.w)
            f.write('qx="%.5f" ' % frame_info.quaternion.x)
            f.write('qy="%.5f" ' % frame_info.quaternion.y)
            f.write('qz="%.5f" ' % frame_info.quaternion.z)
            f.write('s="%.5f" ' % frame_info.scale)
            f.write('/>\n')
                                
            i += 1
            
        f.write(' </array>\n\n')

    f.write('\n</root>\n')
    f.close()
    
    
def save_animation_to_anim(filename, scene, rig_name, action_name, bones, start_frame_number, end_frame_number):
    container = ContainerWriter()
    
    segment = ContainerSegment("INFO")
    segment.append_string_with_null_terminator("version")
    segment.append_uint32(1)
    segment.append_string_with_null_terminator("animation_name")
    segment.append_string_with_null_terminator(action_name)
    segment.append_string_with_null_terminator("frames_per_second")
    segment.append_uint32(int(scene.render.fps / scene.render.fps_base))
    segment.append_string_with_null_terminator("skeleton")
    segment.append_string_with_null_terminator(rig_name)
    container.add_segment(segment)
    
    segment = ContainerSegment("BONES")
    segment.append_uint32(len(bones))
    for pb in bones:
        segment.append_string_with_null_terminator(pb.name)
    container.add_segment(segment)

    segment = ContainerSegment("FRAMES")
    segment.append_string_with_null_terminator("compression_method")
    segment.append_string_with_null_terminator("UNCOMPRESSED")
    num_frames = end_frame_number-start_frame_number
    segment.append_uint32(num_frames)
    i = 0
    for frame in range(start_frame_number, end_frame_number):
        scene.frame_set(frame)
        
        segment.append_uint32(i)
        
        for pb in bones:
            frame_info = FrameInfo(pb)
            
            segment.append_float(frame_info.position.x)
            segment.append_float(frame_info.position.y)
            segment.append_float(frame_info.position.z)
            segment.append_float(frame_info.quaternion.w)
            segment.append_float(frame_info.quaternion.x)
            segment.append_float(frame_info.quaternion.y)
            segment.append_float(frame_info.quaternion.z)
            segment.append_float(frame_info.scale)
                                
        i += 1
        
    container.add_segment(segment)
    
    container.write_to_file(filename)
    

class FrameInfo():
    def __init__(self, bone):
        matrix = get_visual_matrix(bone)
        matrix = scale_matrix(matrix, 60.0, 60.0, 60.0)
        self.quaternion = matrix.to_quaternion()
        self.position = matrix.to_translation()
        self.scale = matrix_to_uniform_scale(matrix)

        
class ContainerSegment():
    def __init__(self, name):
        self.name = name
        self.buffer = bytes()
        
    def get_total_size(self):
        return len(self.name)+1 + len(self.buffer)
        
    def append_float(self, val):
        self.buffer += struct.pack('f', val)

    def append_uint8(self, val):
        self.buffer += struct.pack('B', val)

    def append_sint8(self, val):
        self.buffer += struct.pack('b', val)

    def append_uint16(self, val):
        self.buffer += struct.pack('H', val)

    def append_sint32(self, val):
        self.buffer += struct.pack('i', val)
        
    def append_uint32(self, val):
        self.buffer += struct.pack('I', val)
        
    def append_string_with_null_terminator(self, val):
        if (len(val) > 0):
            self.buffer += bytes(val, encoding)
        self.buffer += struct.pack('!B', 0)


class ContainerWriter():
    def __init__(self):
        self.segments = []
        
    def add_segment(self, segment):
        self.segments.append(segment)
        
    def write_to_file(self, filename):
        file = open(filename, 'wb')
        
        head_segment = ContainerSegment("")
        head_segment.append_string_with_null_terminator("10TONS_CONTAINER")
        head_segment.append_string_with_null_terminator("VERSION")
        head_segment.append_uint32(1)
        
        # saving file size 0 marks undertermined file size (can't do safety checks but the file is easier to generate)
        head_segment.append_string_with_null_terminator("FILE_SIZE")
        head_segment.append_uint32(0)
        
        file.write(head_segment.buffer)

        end_segment = ContainerSegment("")
        end_segment.append_string_with_null_terminator("SEGMENT_END")
        
        for segment in self.segments:
            start_segment = ContainerSegment("")
            start_segment.append_string_with_null_terminator("SEGMENT")
            start_segment.append_uint32(segment.get_total_size()) # how many bytes we need to skip to get to the SEGMENT_END
            start_segment.append_string_with_null_terminator(segment.name)
            start_segment.append_uint32(len(segment.buffer)) # segment data size
            file.write(start_segment.buffer)
            file.write(segment.buffer)
            file.write(end_segment.buffer)
            
            
        # all containers need to end with specific marker
        tail_segment = ContainerSegment("")
        tail_segment.append_string_with_null_terminator("CONTAINER_END")
        file.write(tail_segment.buffer)
        file.close()    
	
    
def get_visual_matrix(pb):

    data_bone = pb.bone
    pose = pb.matrix
    data = pb.bone.matrix_local

    if data_bone.parent:
        data_parent = data_bone.parent.matrix_local
        pose_parent = pb.parent.matrix
    else:
        data_parent = mathutils.Matrix()
        pose_parent = mathutils.Matrix()
        
    visual_matrix = data.inverted() * data_parent * pose_parent.inverted() * pose

    return visual_matrix
	
	
def should_export_bone(pb):
    if (str(pb.name).find(".IK") != -1):
        return False
    if (str(pb.name).find("-trgt") != -1):
        return False
    if (str(pb.name).find("-cntrl") != -1):
        return False
    return pb.bone.use_deform
    
def scale_matrix(matrix, x, y, z):
    matrix[0][3] *= x
    matrix[1][3] *= y
    matrix[2][3] *= z
    
    return matrix
       
	
def matrix_to_uniform_scale(m):
    base_x = mathutils.Vector((m[0][0], m[1][0], m[2][0]))
    base_y = mathutils.Vector((m[0][1], m[1][1], m[2][1]))
    base_z = mathutils.Vector((m[0][2], m[1][2], m[2][2]))
   
    uniform_scale = max(base_x.length, base_y.length, base_z.length)

    return uniform_scale
        
      
def menu_func(self, context):
    self.layout.operator(Export10tonsAnimation.bl_idname, text="10tons Animation (.xml, .anim)")

def register():
    bpy.utils.register_module(__name__)
    bpy.types.INFO_MT_file_export.append(menu_func)

def unregister():
    bpy.utils.unregister_module(__name__)
    bpy.types.INFO_MT_file_export.remove(menu_func)

if __name__ == "__main__":
    register()