bl_info = {"name": "Export 10tons Model",
           "author": "10tons feat Mysterious Pony",
           "version": (0, 3, 4),
           "blender": (2, 73, 0),
           "location":     "File > Import-Export",
           "description": "Exports model to .xml or .model.",
           "category": "Import-Export"}

import bpy
import struct

from bpy_extras.io_utils import ImportHelper, ExportHelper
from bpy.props import EnumProperty, FloatProperty

exporter_name = "10tons Model"
encoding = "iso-8859-1"

class ExitAbort(Exception):
    pass

class Export10tonsModel(bpy.types.Operator, ExportHelper):
    bl_idname = "export_scene.the10tons_model"
    bl_label = "Export " + exporter_name

    filename_ext = ""

    scale = FloatProperty(
            name="Scale",
            min=0.01, max=1000.0,
            default=60.0,
            )

    export_format_items = [("MODEL", ".model", "", 0), ("XML", ".xml", "", 1)]
    export_format_enum = EnumProperty(name="Format", items=export_format_items)

    def invoke(self, context, event):
        scene = context.scene
        objects = list(ob for ob in scene.objects if ob.select)

        for ob in objects:
            self.filepath = ob.name

        if (len(objects) > 1):
            self.filepath = ""

        context.window_manager.fileselect_add(self)
        return {'RUNNING_MODAL'}

    def execute(self, context):
        if (self.properties.export_format_enum == "XML"):
            self.filename_ext = ".xml"
        elif (self.properties.export_format_enum == "MODEL"):
            self.filename_ext = ".model"
        else:
            self.filename_ext = ".unknown"
            return
        try:
            path = self.properties.filepath + self.filename_ext
            result = save(path, context, self.properties.scale, self.properties.export_format_enum)
            return result
        except ExitAbort as e:
            print("Aborting export. Fix the errors shown above.")
            print("Failed:", e)
            return {'FINISHED'}


def save(path, context, scale, export_format):
    print("\n\n10tons Mesh Exporter.")
    import time
    import os
    import mathutils
    from bpy_extras.io_utils import create_derived_objects, free_derived_objects

    if (export_format == "XML"):
        print("Exporting .xml format.")
    elif (export_format == "MODEL"):
        print("Exporting .model format.")
    else:
        print("Unknown export format. Not doing anything.")
        return {'CANCELLED'}

    previous_phase_start_time = time.clock()

    # export selected only if something selected; otherwise export everything
    # fix this, this is rubbish
    scene = context.scene

    #something_selected = any(ob.is_visible(scene) for ob in scene.objects)
    #something_selected = any(ob.select for ob in scene.objects)

    objects = (ob for ob in scene.objects if ob.select)

    nexus_data_objects = []

    for ob in objects:
        # get derived objects, free is true if free_derived_objects should be called
        # derived is a list of duplicated objects??
        free, derived = create_derived_objects(scene, ob)

        if derived is None:
            continue

        for ob_derived, ob_matrix_local in derived:
            # object type is enum in [‘MESH’, ‘CURVE’, ‘SURFACE’, ‘META’, ‘FONT’, ‘ARMATURE’, ‘LATTICE’, ‘EMPTY’, ‘CAMERA’, ‘LAMP’, ‘SPEAKER’], default ‘EMPTY’, (readonly)
            if ob.type not in {'MESH', 'CURVE', 'SURFACE', 'FONT', 'META'}:
                continue

            nexus_data = NexusData()
            nexus_data_objects.append(nexus_data)

            nexus_data.name = ob.name

            was_in_pose_mode = False
            armature = ob.find_armature()
            if armature is not None:
                print("Found armature: " + armature.name)
                # TODO: is there any way to do this without affecting object that is being exported?
                if (armature.data.pose_position == 'POSE'):
                    armature.data.pose_position = 'REST'
                    scene.update()
                    was_in_pose_mode = True

                nexus_data.skeleton.name = armature.name
                for bone in armature.data.bones:
                    if ".IK" in bone.name:
                        print ("Bone '" + bone.name + "' name contains '.IK'. Skipping it.")
                        continue
                    if "-trgt" in bone.name:
                        print ("Bone '" + bone.name + "' name contains '-trgt'. Skipping it.")
                        continue
                    if "-cntrl" in bone.name:
                        print ("Bone '" + bone.name + "' name contains '-cntrl'. Skipping it.")
                        continue
                    if (not bone.use_deform):
                        print ("Bone '" + bone.name + "' doesn't deform. Skipping it. If you want to export this bone, check the deform checkbox in Blender bone settings.")
                        continue
                    if bone.parent is not None:
                        parent_name = bone.parent.name
                    else:
                        parent_name = ""
                    new_bone = Bone(bone.name, parent_name, bone.matrix_local)

                    transform_matrix = mathutils.Matrix()
                    transform_matrix.identity()
                    transform_matrix[0][3] = -ob_derived.location[0]
                    transform_matrix[1][3] = -ob_derived.location[1]
                    transform_matrix[2][3] = -ob_derived.location[2]
                    # TODO: should we just use the full inversed ob_derived transform here?
                    new_bone.matrix = transform_matrix * new_bone.matrix
                    nexus_data.skeleton.bones.append(new_bone)
                    print("Added bone " + bone.name)


            try:
                # to_mesh params: scene within in which to evaluate modifiers, apply modifiers, modifier settings to apply (enum in ['PREVIEW', 'RENDER']) [, calc_tessface (boolean, (optional)) – Calculate Tessellation, Calculate tessellation faces]
                data = ob_derived.to_mesh(scene, True, 'PREVIEW', calc_undeformed=True)
            except:
                data = None

            # could triangulate be done with modifier instead?
            import bmesh
            bm = bmesh.new()
            bm.from_mesh(data)
            bmesh.ops.triangulate(bm, faces=bm.faces)
            bm.to_mesh(data)
            bm.free()
            del bm


            if data:
                # scale ob_matrix_local by user defined scale in export options (do NOT change the actual ob_matrix_local!)
                nexus_scale_matrix = mathutils.Matrix.Scale(scale, 4)
                export_transformation_matrix = nexus_scale_matrix * ob_matrix_local
                # undo translation, we only want rotation and blender scale
                export_transformation_matrix[0][3] = ob_matrix_local[0][3] - ob_derived.location[0] - ob_derived.delta_location[0]
                export_transformation_matrix[1][3] = ob_matrix_local[1][3] - ob_derived.location[1] - ob_derived.delta_location[1]
                export_transformation_matrix[2][3] = ob_matrix_local[2][3] - ob_derived.location[2] - ob_derived.delta_location[2]
                data.transform(export_transformation_matrix)
                data.calc_normals()

                index_dictionaries = {} # for storing index correspondences to efficiently avoid duplicate vertices

                # surface per texture
                for material in data.materials:
                    surface = Surface()
                    nexus_data.surfaces.append(surface)

                    # Old deprecated method for defining materials assumes the first slot is diffuse, and other texture slots are defined as properties.
                    # We still need to support this for backwards compatibility, but the new named material texture definitions will override these during resolve_textures_from_material.
                    if (material.texture_slots[0] != None):
                        nexus_material = Material(material.name, material.texture_slots[0].texture.image.filepath, material.get("glow_map"), material.get("normal_map"), material.get("specular_map"), material.get("tags"))
                    else:
                        nexus_material = Material(material.name, None, material.get("glow_map"), material.get("normal_map"), material.get("specular_map"), material.get("tags"))

                    nexus_material.resolve_textures_from_material(material)
                    nexus_data.materials.append(nexus_material)
                    index_dictionaries[surface] = {}


                # index_dictionaries contains a dictionary per surface,
                # and such dictionary will contain blender vertex indices as keys, and lists of corresponding nexus vertex indices as values
                # (note that nexus vertex index means indices in surface.vertices (not surface.indices which contains blender's tessfaces))
                #
                # Loop through all tessface vertices
                # Check from surface's index dictionary if vertex has corresponding nexus vertex (i.e. coordinates, normal and texture coordinates are equal)
                # If yes, use that
                # If not, create nexus vertex and add it into surface's index dictionary
                data.calc_tessface()
                for tessface in data.tessfaces: # Tessellation faces of the mesh (derived from polygons)
                    surface = nexus_data.surfaces[tessface.material_index]
                    index_dictionary = index_dictionaries[surface]
                    for index in range(len(tessface.vertices) - 1, -1, -1): # has to be looped backwards or will not be drawn correctly in Nexus
                        blender_vertex_index = tessface.vertices[index] # tessface.vertices: Vertex indices, int array
                        blender_vertex = data.vertices[blender_vertex_index]
                        uv_coords = data.tessface_uv_textures[0].data[tessface.index].uv[index]
                        nexus_vertex_exists = False
                        if blender_vertex_index in index_dictionary:
                            for nexus_vertex_index in index_dictionary[blender_vertex_index]:
                                nexus_vertex = surface.vertices[nexus_vertex_index]
                                if (areVectorsEqual(blender_vertex.co, nexus_vertex.position) and
                                  areVectorsEqual(blender_vertex.normal, nexus_vertex.normal) and
                                  areVectorsEqual(uv_coords, nexus_vertex.texture_coordinates)):
                                    surface.indices.append(nexus_vertex_index)
                                    nexus_vertex_exists = True
                                    break
                        if not nexus_vertex_exists:
                            nexus_vertex_index = len(surface.vertices)
                            surface.indices.append(nexus_vertex_index)
                            new_vertex = Vertex(blender_vertex.co, blender_vertex.normal, uv_coords)
                            surface.vertices.append(new_vertex)
                            if blender_vertex_index not in index_dictionary:
                                index_dictionary[blender_vertex_index] = []
                            index_dictionary[blender_vertex_index].append(nexus_vertex_index)

                            # bone weights
                            if len(nexus_data.skeleton.bones) != 0:
                                weight_found = False
                                for vertex_group in blender_vertex.groups: # vertex_group: mesh vertice group contains index and weight
                                    for vg in ob.vertex_groups: # object's vertex_group contains id and name, match current vertex's index to object's vertex groups to find group's name which will be matched to bone name
                                        if vg.index == vertex_group.group:
                                            for bone_index in range(0, len(nexus_data.skeleton.bones)):
                                                bone = nexus_data.skeleton.bones[bone_index]
                                                if vg.name == bone.id:
                                                    vertex_weight = vertex_group.weight
                                                    surface.vertex_weights.append(VertexBoneWeight(nexus_vertex_index, bone_index, vertex_weight)) # vertex_group.weight == vg.weight(blender_vertex_index)
                                                    weight_found = True
                                # all vertices not bound to any bone are automatically bound to root bone
                                if not weight_found:
                                    surface.vertex_weights.append(VertexBoneWeight(nexus_vertex_index, 0, 1))
                            else:
                                # No skeleton or bones. Extract weights for vertices. Each vertex ends up having a weight value which is clamped between 0 and 1. This value is then exported in color channel.
                                for vertex_group in blender_vertex.groups:
                                    for vg in ob.vertex_groups:
                                        if vg.index == vertex_group.group:
                                            surface.vertices[nexus_vertex_index].weight += vertex_group.weight
                                            if (vertex_group.weight > 0):
                                                surface.has_non_bone_vertex_weights = True


                adjust_weights(nexus_data)

            # armature pose_position was changed for export, change it back
            if was_in_pose_mode:
                armature.data.pose_position = 'POSE'
                scene.update()


        if free:
            free_derived_objects(ob)

    print("Data collection time: %.2f" % (time.clock() - previous_phase_start_time))
    previous_phase_start_time = time.clock()

    export_count = len(nexus_data_objects)
    print("Exporting " + str(export_count) + " objects")

    for nexus_data in nexus_data_objects:
        path_directory = path[:path.rfind(os.sep) + 1]

        convert_data_to_nexus_format(nexus_data, scale)

        if (export_format == "XML"):
            if export_count > 1:
                tree_save_path = path_directory + nexus_data.name + ".xml"
            else:
                tree_save_path = path

            print("Exporting " + nexus_data.name + " to " + tree_save_path)

            # prevent double extensions sometimes caused by the blender file selector
            tree_save_path = tree_save_path.replace(".xml.xml", ".xml")
            write_to_file_xml(tree_save_path, nexus_data)

        elif (export_format == "MODEL"):
            if export_count > 1:
                tree_save_path = path_directory + nexus_data.name + ".model"
            else:
                tree_save_path = path

            print("Exporting " + nexus_data.name + " to " + tree_save_path)

            # prevent double extensions sometimes caused by the blender file selector
            tree_save_path = tree_save_path.replace(".model.model", ".model")
            write_to_file_model(tree_save_path, nexus_data)

        print("File save time: %.2f" % (time.clock() - previous_phase_start_time))
        previous_phase_start_time = time.clock()

        copy_textures(path_directory, nexus_data.materials)

        print("Image copying time: %.2f" % (time.clock() - previous_phase_start_time))
        previous_phase_start_time = time.clock()

    return {'FINISHED'}

def copy_texture(map_id, export_source_path, export_target_path, material_subpath_and_filename):
    import shutil
    import os
    import filecmp
    import bpy.path
    if material_subpath_and_filename:
        print("Texture map '" + map_id + "' = " + material_subpath_and_filename)
        full_source_path = export_source_path + material_subpath_and_filename
        full_source_path = full_source_path.replace("/", os.sep)

        full_target_path = export_target_path + material_subpath_and_filename
        full_target_path = full_target_path.replace("/", os.sep)

        print("full_source_path = '" + full_source_path + "'")
        print("full_target_path = '" + full_target_path + "'")

        print("Copying texture " + full_source_path + " to " + full_target_path + "...")

        target_path_without_filename = full_target_path[:full_target_path.rfind(os.sep) + 1]
        if not os.path.exists(target_path_without_filename):
            os.mkdir(target_path_without_filename)

        if (os.path.exists(full_target_path) and filecmp.cmp(full_source_path, full_target_path)):
            print(full_target_path + " exists, not copying image")
        elif not os.path.exists(full_source_path):
            error_text = "Texture " + full_source_path + " not found!"
            print("ERROR:" + error_text)
            raise Exception(error_text)
        else:
            shutil.copy2(full_source_path, full_target_path)

def copy_textures(export_target_path, materials):
    import shutil
    import os
    import filecmp
    import bpy.path

    export_source_path = bpy.path.abspath("//")
    print("Texture source path: " + export_source_path)
    print("Texture target path: " + export_target_path)
    for material in materials:
        if (material.uses_placeholder_textures):
            continue

        copy_texture("diffuse_map", export_source_path, export_target_path, material.diffuse_map)
        if material.glow_map:
            copy_texture("glow_map", export_source_path, export_target_path, material.glow_map)
        if material.normal_map:
            copy_texture("normal_map", export_source_path, export_target_path, material.normal_map)
        if material.specular_map:
            copy_texture("specular_map", export_source_path, export_target_path, material.specular_map)


def adjust_weights(nexus_data):
    for surface in nexus_data.surfaces:
        weights_by_vertex = {}
        for vertex_w in surface.vertex_weights:
            if vertex_w.vertex_index not in weights_by_vertex:
                weights_by_vertex[vertex_w.vertex_index] = vertex_w.weight
            else:
                weights_by_vertex[vertex_w.vertex_index] = weights_by_vertex[vertex_w.vertex_index] + vertex_w.weight
        for vertex_w in surface.vertex_weights:
            vertex_w.weight = vertex_w.weight/weights_by_vertex[vertex_w.vertex_index]


def areVectorsEqual(a, b):
    for i in range(0, len(a)):
        if not areFloatsEqual(a[i], b[i]):
            return False
    return True

def areFloatsEqual(a, b):
    return abs(a - b) < 0.0000001

class NexusData:
    def __init__(self):
        self.name = None
        self.materials = []
        self.skeleton = Skeleton()
        self.surfaces = []

class Skeleton:
    def __init__(self):
        self.name = ""
        self.bones = []

class Bone:
    def __init__(self, id, parent, new_matrix):
        self.id = id
        self.parent = parent
        self.matrix = new_matrix.copy()

class Material:
    def __init__(self, id, diffuse_map, glow_map, normal_map, specular_map, tags):
        self.blender_diffuse_map_path = diffuse_map
        self.texture_base_path = "" # resolved later
        self.id = id
        self.diffuse_map = diffuse_map
        self.glow_map = glow_map
        self.normal_map = normal_map
        self.specular_map = specular_map
        self.tags = tags
        self.uses_placeholder_textures = False
    def resolve_textures_from_material(self, material):

        # If the first slot is undefined, assume this is a placeholder.
        texture_slot = material.texture_slots[0]
        if texture_slot == None or texture_slot.texture == None:
            self.diffuse_map = "models/textures/" + material.name + ".jpg"
            self.uses_placeholder_textures = True

        for texture_slot in material.texture_slots:
            if texture_slot == None:
                continue
            if texture_slot.name == "diffuse_map":
                self.diffuse_map = texture_slot.texture.image.filepath
            if texture_slot.name == "glow_map":
                self.glow_map = texture_slot.texture.image.filepath
            if texture_slot.name == "normal_map":
                self.normal_map = texture_slot.texture.image.filepath
            if texture_slot.name == "specular_map":
                self.specular_map = texture_slot.texture.image.filepath


class Surface:
    def __init__(self):
        self.indices = []               # contains integers  (how vertices form surfaces: three consecutive indices form a triangle; indices refer to indices in vertices list)
        self.vertices = []              # contains Vertex instances
        self.vertex_weights = []        # contains VertexBoneWeights
        self.has_non_bone_vertex_weights = False    # has non-zero weights assigned to each vertex

class Vertex:
    def __init__(self, position, normal, texture_coordinates):
        self.position = position                         # contains (x, y, z)-triplet (vertices)
        self.normal = normal                             # contains (x, y, z)-triplet
        self.texture_coordinates = texture_coordinates   # contains (u, v)-pair
        self.weight = 0                                  # single weight between 0 and 1. Note this is NOT a bone weight.


class VertexBoneWeight:
    def __init__(self, vertex_index, bone_index, weight):
        self.vertex_index = vertex_index
        self.bone_index = bone_index
        self.weight = weight

def get_texture_path_normalized(texture_path):
    processed_path = texture_path.replace("\\", "/")
    processed_path = processed_path.replace("//", "") # remove blender \\ notation in case it was set
    return processed_path

def convert_data_to_nexus_format(nexus_data, scale):
    import re

    nexus_c = NexusCoordinates()

    # remove any materials not mapped to any vertices
    material_index = 0
    for surface in nexus_data.surfaces:
        if (len(surface.indices) == 0):
            del nexus_data.surfaces[material_index]
            print("Removing material '{}' (no vertices referenced).".format(
                nexus_data.materials[material_index].id))

            del nexus_data.materials[material_index]
            material_index = material_index - 1

        material_index = material_index + 1

    # process material attributes to saveable format
    for material in nexus_data.materials:
        if (not material.uses_placeholder_textures):
            print("Material '" + material.id + "' diffuse source path is '" + material.diffuse_map + "'")
            if (not material.diffuse_map.startswith("//")):
                error_text = "ERROR: Incorrectly starting diffuse texture path '" + material.diffuse_map + "'. The path needs to be relative to the .blend file and start with // characters. For example: '//textures/myatlas.jpg'."
                print(error_text)
                raise Exception(error_text)

            material.diffuse_map = re.sub("^[/]*", "", material.diffuse_map) # Remove any '/' characters from the beginning of material path
            material.diffuse_map = material.diffuse_map.replace("\\", "/")
            material.texture_base_path = "./"

            if material.glow_map is not None:
                material.glow_map = get_texture_path_normalized(material.glow_map)
            if material.normal_map is not None:
                material.normal_map = get_texture_path_normalized(material.normal_map)
            if material.specular_map is not None:
                material.specular_map = get_texture_path_normalized(material.specular_map)

    for bone in nexus_data.skeleton.bones:
        original_matrix = bone.matrix.copy()
        for i in range(0, 4):
            for j in range(0, 4):
                if i == 0:
                    value = original_matrix[1][j]
                elif i == 1:
                    value = original_matrix[0][j]
                else:
                    value = original_matrix[i][j]
                if 0 <= i <= 2:
                    if j == 3:
                        value = -scale * value
                    else:
                        value = -value
                bone.matrix[i][j] = value

    if len(nexus_data.skeleton.bones) > 0:
        matrix = nexus_data.skeleton.bones[0].matrix

        for i in range(3):
            if abs(matrix[i][3]) <= 1.0:
                matrix[i][3] = 0.0


"""
    # turns out we can't adjust the vertex data as it would change the model in blender
    # we might want to measure the performance impact of taking copies of everything later


    # map positions, normals, and texture coordinates to Nexus model format
    surface_index = 0
    for surface in nexus_data.surfaces:
        for vertex in surface.vertices:

            old_position = vertex.position
            vertex.position[0] = nexus_c.x(old_position)
            vertex.position[1] = nexus_c.y(old_position)
            vertex.position[2] = nexus_c.z(old_position)

            old_normal = vertex.normal
            vertex.normal[0] = nexus_c.x(old_normal)
            vertex.normal[1] = nexus_c.y(old_normal)
            vertex.normal[2] = nexus_c.z(old_normal)

            old_texture_coordinates = vertex.texture_coordinates
            vertex.texture_coordinates[0] = nexus_c.u(old_texture_coordinates)
            vertex.texture_coordinates[1] = nexus_c.v(old_texture_coordinates)
"""


def write_to_file_xml(path, nexus_data):
    import re
    import os

    nexus_c = NexusCoordinates()

    file = open(path, 'wb')
    writer = XmlWriter(file)
    writer.write("<?xml version=\"1.0\" encoding=\"" + encoding + "\"?>")
    writer.write("<root>\n")

    writer.set_intend_level(1)
    writer.write("<array id=\"INFO\">")
    writer.set_intend_level(2)
    writer.write("<node id=\"format\" version=\"0.9.2\" />")
    writer.write("<node id=\"coordinate_system\" left=\"-x\" up=\"-y\" forward=\"+z\" />")
    writer.set_intend_level(1)
    writer.write("</array>\n")

    writer.write("<array id=\"MATERIALS\">")
    writer.set_intend_level(2)

    for material in nexus_data.materials:
        material_node = "<node id=\"" + material.id + "\""
        if material.tags is not None:
            material_node += " tags=\"" + material.tags + "\""
        material_node += " diffuse_map=\"" + material.texture_base_path + material.diffuse_map + "\""
        if material.glow_map is not None:
            material_node += " glow_map=\"" + material.texture_base_path + material.glow_map + "\""
        if material.normal_map is not None:
            material_node += " normal_map=\"" + material.texture_base_path + material.normal_map + "\""
        if material.specular_map is not None:
            material_node += " specular_map=\"" + material.texture_base_path + material.specular_map + "\""
        material_node += " />"

        writer.write(material_node)

    writer.set_intend_level(1)
    writer.write("</array>\n")

    if len(nexus_data.skeleton.bones) != 0:
        writer.write("<array id=\"SKELETON\" name=\"" + nexus_data.skeleton.name + "\">")
        writer.set_intend_level(2)
        for bone in nexus_data.skeleton.bones:
            bone_node = "<node id=\"" + str(bone.id) + "\""
            if bone.parent is not None and bone.parent is not "":
                bone_node += " parent=\"" + bone.parent + "\""
            writer.write(bone_node)
            writer.set_intend_level(3)
            for i in range(0, 4):
                line = ""
                for j in range(0, 4):
                    value = round(bone.matrix[i][j], 4)
                    line += "m" + str(i+1) + str(j + 1) + "=\"" + str(value) + "\" "
                writer.write(line)
            writer.set_intend_level(2)
            writer.write("/>")
        writer.set_intend_level(1)
        writer.write("</array>\n")


    writer.write("<array id=\"SURFACES\">")
    writer.set_intend_level(2)
    surface_index = 0
    for surface in nexus_data.surfaces:
        material_id = nexus_data.materials[surface_index].id
        writer.write("<node id=\"" + str(surface_index) + "\" primitive_format=\"TRIANGLES\" material=\"" + material_id + "\" />" )
        surface_index = surface_index + 1
    writer.set_intend_level(1)
    writer.write("</array>\n")

    surface_index = 0
    for surface in nexus_data.surfaces:
        writer.write("<array id=\"SURFACE_" + str(surface_index) + "_INDICES\">")
        writer.set_intend_level(2)
        for index in surface.indices:
            writer.write("<node i=\"" + str(index) + "\"/>")
        writer.set_intend_level(1)
        writer.write("</array>\n")

        writer.write("<array id=\"SURFACE_" + str(surface_index) + "_POSITIONS\">")
        writer.set_intend_level(2)
        for vertex in surface.vertices:
            writer.write("<node x=\"" + str(nexus_c.x(vertex.position)) + "\" y=\"" + str(nexus_c.y(vertex.position)) + "\" z=\"" + str(nexus_c.z(vertex.position)) + "\" />")
        writer.set_intend_level(1)
        writer.write("</array>\n")

        if (surface.has_non_bone_vertex_weights):
            # encode vertex weights into the alpha channel
            writer.write("<array id=\"SURFACE_" + str(surface_index) + "_COLORS\">")
            for vertex in surface.vertices:
                writer.write("<node r=\"1\" g=\"1\" b=\"1\" a=\"" + str(round(vertex.weight, 4)) + "\" />")
            writer.set_intend_level(1)
            writer.write("</array>\n")


        writer.write("<array id=\"SURFACE_" + str(surface_index) + "_NORMALS\">")
        for vertex in surface.vertices:
            writer.write("<node x=\"" + str(nexus_c.x(vertex.normal)) + "\" y=\"" + str(nexus_c.y(vertex.normal)) + "\" z=\"" + str(nexus_c.z(vertex.normal)) + "\" />")
        writer.set_intend_level(1)
        writer.write("</array>\n")

        writer.write("<array id=\"SURFACE_" + str(surface_index) + "_TEXTURE_COORDINATES_0\">")
        for vertex in surface.vertices:
            writer.write("<node u=\"" + str(nexus_c.u(vertex.texture_coordinates)) + "\" v=\"" + str(nexus_c.v(vertex.texture_coordinates)) + "\" />")
        writer.set_intend_level(1)
        writer.write("</array>\n")

        if len(nexus_data.skeleton.bones) != 0:
            writer.write("<array id=\"SURFACE_" + str(surface_index) + "_VERTEX_WEIGHTS\">")
            for vertex_weight in surface.vertex_weights:
                writer.write("<node vi=\"" + str(vertex_weight.vertex_index) + "\" bi=\"" + str(vertex_weight.bone_index) + "\" w=\"" + str(round(vertex_weight.weight, 4)) + "\" />")
            writer.set_intend_level(1)
            writer.write("</array>\n")

        surface_index = surface_index + 1

    writer.set_intend_level(0)
    writer.write("</root>")
    file.close()


def write_to_file_model(path, nexus_data):
    import re
    import os

    nexus_c = NexusCoordinates()

    container = ContainerWriter()

    segment = ContainerSegment("INFO")
    segment.append_string_with_null_terminator("VERSION")
    segment.append_uint32(2)
    container.add_segment(segment)

    segment = ContainerSegment("MATERIALS")
    segment.append_uint32(len(nexus_data.materials))
    for material in nexus_data.materials:
        segment.append_string_with_null_terminator(material.id)

        # tags string
        if material.tags is not None:
            segment.append_string_with_null_terminator(material.tags)
        else:
            segment.append_string_with_null_terminator("")

        num_texture_slots = 0
        if material.diffuse_map is not None:
            num_texture_slots += 1
        if material.glow_map is not None:
            num_texture_slots += 1
        if material.normal_map is not None:
            num_texture_slots += 1
        if material.specular_map is not None:
            num_texture_slots += 1
        segment.append_uint32(num_texture_slots)

        if material.diffuse_map is not None:
            segment.append_string_with_null_terminator("diffuse_map")
            segment.append_string_with_null_terminator(material.texture_base_path + material.diffuse_map)
        if material.glow_map is not None:
            segment.append_string_with_null_terminator("glow_map")
            segment.append_string_with_null_terminator(material.texture_base_path + material.glow_map)
        if material.normal_map is not None:
            segment.append_string_with_null_terminator("normal_map")
            segment.append_string_with_null_terminator(material.texture_base_path + material.normal_map)
        if material.specular_map is not None:
            segment.append_string_with_null_terminator("specular_map")
            segment.append_string_with_null_terminator(material.texture_base_path + material.specular_map)

    container.add_segment(segment)

    if len(nexus_data.skeleton.bones) != 0:
        segment = ContainerSegment("SKELETON")
        container.add_segment(segment)

        segment.append_string_with_null_terminator(nexus_data.skeleton.name)
        segment.append_uint32(len(nexus_data.skeleton.bones))

        for bone in nexus_data.skeleton.bones:
            segment.append_string_with_null_terminator(str(bone.id))

            if bone.parent is not None and bone.parent is not "":
                segment.append_string_with_null_terminator(bone.parent)
            else:
                segment.append_string_with_null_terminator("")

            for i in range(0, 4):
                for j in range(0, 4):
                    segment.append_float(bone.matrix[i][j])


    segment = ContainerSegment("SURFACES")
    container.add_segment(segment)
    surface_index = 0

    segment.append_uint32(len(nexus_data.surfaces))
    for surface in nexus_data.surfaces:
        segment.append_string_with_null_terminator(str(surface_index)) # surface id
        material_id = nexus_data.materials[surface_index].id
        segment.append_string_with_null_terminator(material_id)

        segment.append_uint32(len(surface.vertices)) # number of vertices
        segment.append_uint32(len(surface.indices)) # number of indices
        segment.append_uint32(len(surface.vertex_weights)) # number of vertex weights

        surface_index = surface_index + 1


    surface_index = 0
    for surface in nexus_data.surfaces:

        segment = ContainerSegment("SURFACE_" + str(surface_index) + "_INDICES")
        container.add_segment(segment)
        for index in surface.indices:
            segment.append_uint16(index)

        segment_positions = ContainerSegment("SURFACE_" + str(surface_index) + "_POSITIONS")

        segment_colors = None
        if (surface.has_non_bone_vertex_weights):
            segment_colors = ContainerSegment("SURFACE_" + str(surface_index) + "_COLORS")
        segment_normals = ContainerSegment("SURFACE_" + str(surface_index) + "_NORMALS")
        segment_texture_coordinates = ContainerSegment("SURFACE_" + str(surface_index) + "_TEXTURE_COORDINATES_0")

        container.add_segment(segment_positions)
        if (segment_colors != None):
            container.add_segment(segment_colors)
        container.add_segment(segment_normals)
        container.add_segment(segment_texture_coordinates)

        for vertex in surface.vertices:
            segment_positions.append_float(nexus_c.x(vertex.position))
            segment_positions.append_float(nexus_c.y(vertex.position))
            segment_positions.append_float(nexus_c.z(vertex.position))
            segment_normals.append_float(nexus_c.x(vertex.normal))
            segment_normals.append_float(nexus_c.y(vertex.normal))
            segment_normals.append_float(nexus_c.z(vertex.normal))
            segment_texture_coordinates.append_float(nexus_c.u(vertex.texture_coordinates))
            segment_texture_coordinates.append_float(nexus_c.v(vertex.texture_coordinates))

        # Export non-bone vertex weights encoded in color alpha channel.
        if (surface.has_non_bone_vertex_weights):
            for vertex in surface.vertices:
                segment_colors.append_uint8(255) # r
                segment_colors.append_uint8(255) # g
                segment_colors.append_uint8(255) # b
                segment_colors.append_uint8(int(round(255*vertex.weight, 0))) # a


        if len(nexus_data.skeleton.bones) != 0:
            segment = ContainerSegment("SURFACE_" + str(surface_index) + "_VERTEX_WEIGHTS")
            container.add_segment(segment)

            for vertex_weight in surface.vertex_weights:
                segment.append_uint16(vertex_weight.bone_index)
                segment.append_uint16(vertex_weight.vertex_index)
                segment.append_float(vertex_weight.weight)

        surface_index = surface_index + 1


    container.write_to_file(path)


class NexusCoordinates():
    def x(self, vector):
        return round( -vector[1], 5)

    def y(self, vector):
        return round( -vector[0], 5)

    def z(self, vector):
        return round( -vector[2], 5)

    def u(self, vector):
        return vector[0]

    def v(self, vector):
        return 1 - vector[1]

class XmlWriter():
    def __init__(self, file):
        self.file = file
        self.intend_level = 0

    def write(self, line):
        intend = ''.join([' ' for s in range(self.intend_level * 2)])
        self.file.write(bytes(intend + line + "\n", encoding))

    def set_intend_level(self, level):
        self.intend_level = level

class ContainerSegment():
    def __init__(self, name):
        self.name = name
        self.buffer = bytes()

    def get_total_size(self):
        return len(self.name)+1 + len(self.buffer)

    def append_float(self, val):
        self.buffer += struct.pack('f', val)

    def append_uint8(self, val):
        self.buffer += struct.pack('B', val)

    def append_sint8(self, val):
        self.buffer += struct.pack('b', val)

    def append_uint16(self, val):
        self.buffer += struct.pack('H', val)

    def append_sint32(self, val):
        self.buffer += struct.pack('i', val)

    def append_uint32(self, val):
        self.buffer += struct.pack('I', val)

    def append_string_with_null_terminator(self, val):
        if (len(val) > 0):
            self.buffer += bytes(val, encoding)
        self.buffer += struct.pack('!B', 0)


class ContainerWriter():
    def __init__(self):
        self.segments = []

    def add_segment(self, segment):
        self.segments.append(segment)

    def write_to_file(self, filename):
        file = open(filename, 'wb')

        head_segment = ContainerSegment("")
        head_segment.append_string_with_null_terminator("10TONS_CONTAINER")
        head_segment.append_string_with_null_terminator("VERSION")
        head_segment.append_uint32(1)

        # saving file size 0 marks undertermined file size (can't do safety checks but the file is easier to generate)
        head_segment.append_string_with_null_terminator("FILE_SIZE")
        head_segment.append_uint32(0)

        file.write(head_segment.buffer)

        end_segment = ContainerSegment("")
        end_segment.append_string_with_null_terminator("SEGMENT_END")

        for segment in self.segments:
            start_segment = ContainerSegment("")
            start_segment.append_string_with_null_terminator("SEGMENT")
            start_segment.append_uint32(segment.get_total_size()) # how many bytes we need to skip to get to the SEGMENT_END
            start_segment.append_string_with_null_terminator(segment.name)
            start_segment.append_uint32(len(segment.buffer)) # segment data size
            file.write(start_segment.buffer)
            file.write(segment.buffer)
            file.write(end_segment.buffer)


        # all containers need to end with specific marker
        tail_segment = ContainerSegment("")
        tail_segment.append_string_with_null_terminator("CONTAINER_END")
        file.write(tail_segment.buffer)
        file.close()



def menu_func_export(self, context):
    self.layout.operator(Export10tonsModel.bl_idname, text=exporter_name+" (.xml, .model)")

def register():
    print("Registering 10tons exporter.")
    bpy.utils.register_module(__name__)
    bpy.types.INFO_MT_file_export.append(menu_func_export)

def unregister():
    bpy.utils.unregister_module(__name__)
    bpy.types.INFO_MT_file_export.remove(menu_func_export)

if __name__ == "__main__":
    register()

